package java.lang.ref;

import com.ibm.oti.vm.VM;
import com.ibm.tenant.ParameterizedRunnable;
import com.ibm.tenant.TenantContext;
import com.ibm.tenant.TenantException;
import com.ibm.tenant.TenantGlobals;
import com.ibm.tenant.internal.TenantAccessFactory;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2014  All Rights Reserved
 */

/**
 * Abstract class which describes behavior common to all reference objects.
 *
 * @author		OTI
 * @version		initial
 * @since		1.2
 */
public abstract class Reference<T> extends Object {
	private static final int STATE_INITIAL = 0;
	private static final int STATE_CLEARED = 1;
	private static final int STATE_ENQUEUED = 2;

	private T referent;
	private ReferenceQueue queue;
	private int state;

/**
 * Make the referent null.  This does not force the reference object to be enqueued.
 */
public void clear() {
	synchronized(this) {
		referent = null;
		/* change the state to cleared if it's not already cleared or enqueued */
		if (STATE_INITIAL == state) {
			state = STATE_CLEARED;
		}
	}
}

/**
 * Force the reference object to be enqueued if it has been associated with a queue.
 *
 * @return	true if Reference is enqueued, false otherwise.
 */
public boolean enqueue() {
	return enqueueImpl();
}

/**
 * Return the referent of the reference object.
 *
 * @return	the referent to which reference refers,
 *			or null if object has been cleared.
 */
public T get() {
	if (VM.J9_GC_POLICY != VM.J9_GC_POLICY_METRONOME) {
		return referent;
	}
	return getImpl();
}

private native T getImpl();

/**
 * Return whether the reference object has been enqueued.
 *
 * @return	true if Reference has been enqueued, false otherwise.
 */
public boolean isEnqueued () {
	synchronized(this) {
		return state == STATE_ENQUEUED;
	}
}

/**
 * Enqueue the reference object on the associated queue.
 *
 * @return	true if the Reference was successfully
 *			enqueued, false otherwise.
 */
boolean enqueueImpl() {
	final ReferenceQueue tempQueue;
	boolean result;
	T tempReferent = referent;
	synchronized(this) {
		/* Static order for the following code (DO NOT CHANGE) */
		tempQueue = queue;
		queue = null;
		if (state == STATE_ENQUEUED || tempQueue == null) {
			return false;
		}
		if (TenantGlobals.isMTJCLEnabled()) {
			/*
			 * Because enqueueImpl() might be called by the finalizer thread, we need to make ensure this is run in the correct tenant to
			 * avoid root-context->tenant-context references. Switching to the tenant of the ReferenceQueue not the Reference because it
			 * is ReferenceQueue.enqueue() that may allocate objects
			 */
			TenantContext tenant = TenantContext.getTenantForObject(tempQueue);
			if (null == tenant) {
				/* An object's tenant may be null because it is in ROOT owned memory
				 * or because the tenant is shutting down and we are past the point
				 * when the GC deletes its reference to the TenantContext object
				 */
				if (TenantAccessFactory.getAccess().isObjectInRootTenant(tempQueue)) {
					result = tempQueue.enqueue(this);
				} else {
					/* If the object is not in the ROOT and the tenant that owns
					 * the queue is null there is no point in enqueuing the reference
					 * since it will never be dequeued because all of the tenant's
					 * threads will have terminated
					 */
					result = false;
				}
			} else {
				try {
					result = TenantAccessFactory.getAccess().primitiveRun(tenant, new ParameterizedRunnable<Boolean, Void>() {
						public Boolean run(Void runnableData) {
							return new Boolean(tempQueue.enqueue(Reference.this));
						}
					} , null).booleanValue();
				} catch (TenantException e) {
					result = false;
				}
			}
		} else {
			result = tempQueue.enqueue(this);
		}

		if (result) {
			state = STATE_ENQUEUED;
			if (null != tempReferent) {
				reprocess();
			}
		}
		return result;
	}
}

private native void reprocess();

/**
 * Constructs a new instance of this class.
 */
Reference() {
}

/**
 * Initialize a newly created reference object. Associate the
 * reference object with the referent.
 *
 * @param r the referent
 */
void initReference (T r) {
	state = STATE_INITIAL;
	referent = r;
}

/**
 * Initialize a newly created reference object.  Associate the
 * reference object with the referent, and the specified ReferenceQueue.
 *
 * @param r the referent
 * @param q the ReferenceQueue
 */
void initReference (T r, ReferenceQueue q) {
	queue = q;
	state = STATE_INITIAL;
	referent = r;
}

/**
 * Called when a Reference has been removed from its ReferenceQueue.
 * Set the enqueued field to false.
 */
void dequeue() {
	synchronized(this) {
		state = STATE_CLEARED;
	}
}
}
