/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 2001, 2014 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
package javax.rcm;

import com.ibm.oti.vm.VM;

import java.lang.reflect.Method;
import java.security.AccessController;
import java.security.PrivilegedAction;

import com.ibm.rcm.resolver.ResourceConsumerResolver;
import com.ibm.rcm.resolver.ResourceConsumerResolverCenter;
import com.ibm.rcm.resource.CpuTimeResourceAttributes;
import com.ibm.rcm.resource.PlatformResourceAttributes;
import com.ibm.tenant.TenantContext;
import com.ibm.tenant.TenantContextEvent;
import com.ibm.tenant.TenantContextListener;
import com.ibm.tenant.TenantException;
import com.ibm.tenant.TenantGlobalListener;
import com.ibm.tenant.internal.TenantAccessFactory;

/**
 *	A thread controlled by RCM CPU throttling will receive a signal periodically.
 *	Each time, the thread will check whether it is time to report the CPU usage.
 *	If yes, the thread need to calculate the CPU time used since last report.
 *	Then the thread will wait for RCM's permission to continue execution.
 *
 *	This class provides runnable that will be called when threads received the signal.
 *	It also implements TenantContextListener to provide the CPU throttling during tenant attach/detach.
 */
class CPUThrottlingRunnable implements Runnable, TenantContextListener, TenantGlobalListener  {

	/**
	 *	Get the operating system.
	 *
	 *	The timing function for each operating system is different. So we need this information.
	 *	(It is ugly considering the portability though :-( ).
	 */
	enum OS{
		UNKONWN,WIN,LINUX;
		static OS getOS(){
			String osname = System.getProperty("os.name"); //$NON-NLS-1$
			if(null == osname){
				return UNKONWN;
			} else if (osname.contains("Windows")) { //$NON-NLS-1$
				return WIN;
			} else if(osname.contains("Linux")) { //$NON-NLS-1$
				return LINUX;
			} else {
				return UNKONWN;
			}
		}
	}

	/**
	 *	Default impl to get time information. Could be inherited for different OS.
	 */
	private static class DefaultTimer {
		DefaultTimer() {
		}
		long getElapsedTime(){
			return System.currentTimeMillis() * 1000000;
		}
		long getThreadTime(){
			return VM.updateThreadCPUTime();
		}
	}

	private CPUThrottlingRunnable() {
	}

	private static class Holder{
		private static final CPUThrottlingRunnable instance = new CPUThrottlingRunnable();
	}

	/**
	 * @return the singleton instance of this class.
	 */
	static CPUThrottlingRunnable getInstance(){
		return Holder.instance;
	}

	private ThreadLocal<Long> threadUsageHistory = new ThreadLocal<Long>();
	private ThreadLocal<Long> totalUsageHistory = new ThreadLocal<Long>();
	private ThreadLocal<Boolean> isRunning = new ThreadLocal<Boolean>();

	private static long   tokenBucketLimit        = -1;
	private static long   tokenBucketInterval     = -1;
	private static double numberOfProcessingUnits = 0;

	private static DefaultTimer timer = null;

	static {
		CpuTimeResourceAttributes cpu = (CpuTimeResourceAttributes)ResourceAttributesFactory.getInstanceByName(CpuTimeResourceAttributes.class.getName());
		if(null != cpu) {
			numberOfProcessingUnits = cpu.getNumOfLogicalProcessors();
		}
		if (OS.WIN == OS.getOS()) {
			timer = new DefaultTimer();
		} else if(OS.LINUX == OS.getOS()) {
			timer = new DefaultTimer(){
				@Override
				long getElapsedTime(){
					return System.nanoTime();
				}
			};
		} else {
			timer = new DefaultTimer();
		}
	}

	private boolean initVars(PlatformResourceAttributes cpu){
		if(tokenBucketLimit < 0 ) {
			synchronized (CPUThrottlingRunnable.class) {
				if(tokenBucketLimit < 0 ){
				    /* Need to convert the bucket capacity used by this JVM process to be based on the capacity of the whole system,
				     * as RCM limits CPU consumption against the 'maxBandwidth' configured in rcm.xml.
				     */
					tokenBucketLimit = getTokenBucketLimit(cpu.getHandle())*100/cpu.getMaxBandwidth();
					tokenBucketInterval = getTokenBucketInterval(cpu.getHandle()) * 1000000; //milli -> nano
				}
			}
		}
		return tokenBucketLimit >=0;
	}

	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		runThrottling(true);
	}

   /**
    * Runs the throttling for the thread.
    * @param bIsThrottlingRequired indicates if or not the real throttling happened.
    */
	private void runThrottling(final boolean bIsThrottlingRequired) {
		//prevent from re-enter
		Boolean running = isRunning.get();
		if(null != running && running.booleanValue()) {
			return;
		} else {
			isRunning.set(Boolean.TRUE);
			AccessController.doPrivileged(new PrivilegedAction<Void>() {
				public Void run() {
					// throttling for current thread
					ResourceConsumer consumer = resolve(null);
					throttling(consumer, bIsThrottlingRequired);
					return null;
				}
			});
			isRunning.set(Boolean.FALSE);
		}
	}

	/**
	 * Get the {@link javax.rcm.ResourceConsumer} according to the consumer object.
	 * If consumer object is null, use the consumer object in current executing context.
	 *
	 * @param consumerObj the consumer object
	 * @return related ResourceConsumer or NULL if not found
	 */
	private ResourceConsumer resolve(Object consumerObj){
		ResourceConsumerResolver resolver = ResourceConsumerResolverCenter.getResolver();
		if(null == resolver) {
			return null;
		}

		ResourceConsumer consumer = null;
		if(null == consumerObj) {
			consumer = resolver.resolveCurrentResourceConsumer();
		} else if (resolver.isConsumerTypeSupported(consumerObj)) {
			consumer = resolver.resolveResourceConsumer(consumerObj);
		}
		return consumer;
	}

	/* (non-Javadoc)
	 * @see com.ibm.tenant.TenantContextListener#tenantContextDestroyed(com.ibm.tenant.TenantContextEvent)
	 */
	public void tenantContextDestroyed(TenantContextEvent event) {
		//do nothing
	}

	/**
	 * Record the CPU time when the thread is attached to a tenant.
	 *
	 * @see com.ibm.tenant.TenantContextListener#tenantContextAttached(com.ibm.tenant.TenantContextEvent)
	 */
	@SuppressWarnings("boxing")
	public void tenantContextAttached(TenantContextEvent event) {
		threadUsageHistory.set(timer.getThreadTime());
		totalUsageHistory.set(timer.getElapsedTime());
	}

	/**
	 * Make sure that all the CPU time is throttled before the thread detached from tenant.
	 *
	 * @see com.ibm.tenant.TenantContextListener#tenantContextDetached(com.ibm.tenant.TenantContextEvent)
	 */
	public void tenantContextDetached(TenantContextEvent event) {
		TenantContext tc = event.getTenantContext();
		ResourceConsumer consumer = resolve(tc);

		//prevent from re-enter
		while(true){
			Boolean running = isRunning.get();
			if(null != running && running.booleanValue()){
				try {
					wait(1000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}else{
				break;
			}
		}

		isRunning.set(Boolean.TRUE);
		throttling(consumer, true);
		isRunning.set(Boolean.FALSE);
	}

	private static Method applyConstraints;
	private static Method sendNotifications;
	static {
		try {
			applyConstraints = ResourceDomain.class.getDeclaredMethod("applyConstraints", long.class, long.class); //$NON-NLS-1$
			applyConstraints.setAccessible(true);
			sendNotifications = ResourceDomain.class.getDeclaredMethod("sendNotifications", long.class, long.class); //$NON-NLS-1$
			sendNotifications.setAccessible(true);
		} catch (Exception e) {
			//should never happen
			throw new AssertionError("Error during get reflection methods from ResourceDomain"); //$NON-NLS-1$
		}
	}

	/**
	 * Doing the throttling for a {@link javax.rcm.ResourceConsumer}
	 * Customer constraints will be applied before throttling, and notification will be sent after throttling.
	 * @param bIsThrottlingRequired indicates if or not the real throttling happened.
	 * @param consumer the ResourceConsumer
	 */
	@SuppressWarnings("boxing")
	private void throttling(ResourceConsumer consumer, boolean bIsThrottlingRequired ){
		// this must be collected here because as a side effect it updates the amount of time used by the thread
		// we this time to calculate how much time a tenant has used.  If it is done later down we may
		// return before the time is updated.
		long currentThreadUsage = timer.getThreadTime();

		if(null == consumer){
			return;
		}

		PlatformResourceAttributes cpu = (PlatformResourceAttributes)ResourceAttributesFactory.getInstanceByName(CpuTimeResourceAttributes.class.getName());
		if(null == cpu){
			return;
		}

		if(!initVars(cpu)){
			return;
		}

		ResourceDomain domain = consumer.currentDomain(cpu);
		if(null == domain){
			return;
		}

		Long l = threadUsageHistory.get();
		long lastThreadUsage = (null==l)?0:l.longValue();
		l = totalUsageHistory.get();
		long lastTotal = (null==l)?0:l.longValue();
		long currentTotal = timer.getElapsedTime();
		long deltaTotal = currentTotal - lastTotal;

		long countedUsage = lastThreadUsage;

		while(currentThreadUsage > countedUsage){
			//Constraint
			try {
				long constrainted = (Long)applyConstraints.invoke(domain, countedUsage, currentThreadUsage);
				long tokenNumber = toTokenNumber(constrainted - countedUsage, deltaTotal);
				//token bucket && record usage
				long allowed;
				if(bIsThrottlingRequired) {
					allowed = requestToken(tokenNumber);
				} else {
					allowed = tokenNumber;
				}
				countedUsage += toCpuTime(allowed, deltaTotal);
			} catch (Exception e) {
				//should never happen
				throw new AssertionError("Error during invcation reflection method ResourceDomain.applyConstraints"); //$NON-NLS-1$
			}
		}
		threadUsageHistory.set(Long.valueOf(countedUsage));
		totalUsageHistory.set(Long.valueOf(currentTotal));

		//notification
		try {
			sendNotifications.invoke(domain,lastThreadUsage, currentThreadUsage);
		} catch (Exception e) {
			//should never happen
			throw new AssertionError("Error during invcation reflection method ResourceDomain.sendNotifications"); //$NON-NLS-1$
		}
	}

	/**
	 * Translate the CPU usage to token number.
	 *
	 * @param threadUsage 	thread CPU usage
	 * @param totalUsage	elapsed time
	 * @return
	 */
	private static long toTokenNumber(long threadUsage, long totalUsage){
		double percentage = ((double)threadUsage) / (tokenBucketInterval * numberOfProcessingUnits);
		return moreThanRound(tokenBucketLimit * percentage);
	}

	/**
	 * Transalte the token number to CPU time.
	 *
	 * @param tokenNumber	token number
	 * @param totalUsage	elapsed time
	 * @return
	 */
	private static long toCpuTime(long tokenNumber, long totalUsage){
		double percentage = ((double)tokenNumber) / tokenBucketLimit ;
		return moreThanRound(tokenBucketInterval * numberOfProcessingUnits * percentage);
	}

	//TODO mismatch is possible here
	private static long moreThanRound(double doubleNum){
		long longRs = Math.round(doubleNum);
		if(((double)longRs) < doubleNum){
			longRs += 1;
		}
		return longRs;
	}

	private native long requestToken(long token);
	private static native long getTokenBucketLimit(long resourceHandle);
	private static native long getTokenBucketInterval(long resourceHandle);

	/* (non-Javadoc)
	 * @see com.ibm.tenant.TenantContextListener#tenantContextAboutToAttach(com.ibm.tenant.TenantContextEvent, java.lang.Thread)
	 */
	public void tenantContextAboutToAttach(TenantContextEvent event, Thread thread) throws TenantException {
		// Do nothing
	}

	/* (non-Javadoc)
	 * @see com.ibm.tenant.TenantContextListener#tenantContextAttachAborted(com.ibm.tenant.TenantContextEvent, java.lang.Thread)
	 */
	public void tenantContextAttachAborted(TenantContextEvent event, Thread thread) {
		// Do nothing
	}

	/* (non-Javadoc)
	 * @see com.ibm.tenant.TenantGlobalListener#tenantGlobalCreated(TenantContextEvent event))
	 */
	public void tenantGlobalCreated(TenantContextEvent event)
			throws TenantException {
		TenantContext tc = event.getTenantContext();

		/* we want to initialize classes referred by injected runnable as part of tenant creation
		 * so that we needn't hold any class initialization lock during injected code run which possibly leads to deadlock.
		 * */
		TenantAccessFactory.getAccess().primitiveRun(tc, new Runnable() {
			public void run() {
				CPUThrottlingRunnable.this.runThrottling(false);
			}
		});
	}

	/* (non-Javadoc)
	 * @see com.ibm.tenant.TenantGlobalListener#tenantGlobalDestroyed(TenantContextEvent event))
	 */
	public void tenantGlobalDestroyed(TenantContextEvent event)
			throws TenantException {
		// do nothing.

	}
}
